@@ -340,6 +340,32 @@ static void vfsDatabaseDestroy(struct vfsDatabase *d)
 	sqlite3_free(d);
 }
 
+/*
+ * Comment copied entirely for sqlite source code, it is safe to assume
+ * the value 0x40000000 will never change. dq_sqlite_pending_byte is global
+ * to be able to adapt it in the unittest, the value must never be changed.
+ *
+ * ==BEGIN COPY==
+ * The value of the "pending" byte must be 0x40000000 (1 byte past the
+ * 1-gibabyte boundary) in a compatible database.  SQLite never uses
+ * the database page that contains the pending byte.  It never attempts
+ * to read or write that page.  The pending byte page is set aside
+ * for use by the VFS layers as space for managing file locks.
+ *
+ * During testing, it is often desirable to move the pending byte to
+ * a different position in the file.  This allows code that has to
+ * deal with the pending byte to run on files that are much smaller
+ * than 1 GiB.  The sqlite3_test_control() interface can be used to
+ * move the pending byte.
+ *
+ * IMPORTANT:  Changing the pending byte to any value other than
+ * 0x40000000 results in an incompatible database file format!
+ * Changing the pending byte during operation will result in undefined
+ * and incorrect behavior.
+ * ==END COPY==
+ */
+unsigned dq_sqlite_pending_byte = 0x40000000;
+
 /* Get a page from the given database, possibly creating a new one. */
 static int vfsDatabaseGetPage(struct vfsDatabase *d,
 			      uint32_t page_size,
@@ -351,44 +377,57 @@ static int vfsDatabaseGetPage(struct vfsDatabase *d,
 	assert(d != NULL);
 	assert(pgno > 0);
 
-	/* SQLite should access pages progressively, without jumping more than
-	 * one page after the end. */
-	if (pgno > d->n_pages + 1) {
-		rc = SQLITE_IOERR_WRITE;
-		goto err;
-	}
+        /* SQLite should access pages progressively, without jumping more than
+         * one page after the end unless one would attempt to access a page at
+         * `sqlite_pending_byte` offset, skipping a page is permitted then. */
+        bool pending_byte_page_reached = (page_size * d->n_pages == dq_sqlite_pending_byte);
+        if ((pgno > d->n_pages + 1) && !pending_byte_page_reached) {
+                rc = SQLITE_IOERR_WRITE;
+                goto err;
+        }
 
-	if (pgno == d->n_pages + 1) {
-		/* Create a new page, grow the page array, and append the
-		 * new page to it. */
-		void **pages; /* New page array. */
-
-		*page = sqlite3_malloc64(page_size);
-		if (*page == NULL) {
-			rc = SQLITE_NOMEM;
-			goto err;
-		}
-
-		pages = sqlite3_realloc64(d->pages, sizeof *pages * pgno);
-		if (pages == NULL) {
-			rc = SQLITE_NOMEM;
-			goto err_after_vfs_page_create;
-		}
-
-		/* Append the new page to the new page array. */
-		*(pages + pgno - 1) = *page;
-
-		/* Update the page array. */
-		d->pages = pages;
-		d->n_pages = pgno;
-	} else {
+	if (pgno <= d->n_pages) {
 		/* Return the existing page. */
 		assert(d->pages != NULL);
-		*page = d->pages[pgno - 1];
-	}
+		*page = d->pages[pgno-1];
+                return SQLITE_OK;
+        }
+
+        /* Create a new page, grow the page array, and append the
+         * new page to it. */
+        *page = sqlite3_malloc64(page_size);
+        if (*page == NULL) {
+                rc = SQLITE_NOMEM;
+                goto err;
+        }
+
+        void **pages = sqlite3_realloc64(d->pages, sizeof *pages * pgno);
+        if (pages == NULL) {
+                rc = SQLITE_NOMEM;
+                goto err_after_vfs_page_create;
+        }
+
+        pages[pgno-1] = *page;
+
+        /* Allocate a page to store the pending_byte */
+        if (pending_byte_page_reached) {
+                void *pending_byte_page = sqlite3_malloc64(page_size);
+                if (pending_byte_page == NULL) {
+                        rc = SQLITE_NOMEM;
+                        goto err_after_pending_byte_page;
+                }
+                pages[d->n_pages] = pending_byte_page;
+        }
+
+        /* Update the page array. */
+        d->pages = pages;
+        d->n_pages = pgno;
 
 	return SQLITE_OK;
 
+err_after_pending_byte_page:
+        d->pages = pages;
+
 err_after_vfs_page_create:
 	sqlite3_free(*page);
 err:
@@ -1542,7 +1581,7 @@ static int vfsShmUnlock(struct vfsShm *s, int ofst, int n, int flags)
 	}
 
 	for (i = ofst; i < ofst + n; i++) {
-		/* Sanity check that no lock of the other type is held in this
+		/* Coherence check that no lock of the other type is held in this
 		 * region. */
 		assert(other_locks[i] == 0);
 
@@ -2371,6 +2410,20 @@ int VfsAbort(sqlite3_vfs *vfs, const char *filename)
 	return 0;
 }
 
+/* Extract the number of pages field from the database header. */
+static uint32_t vfsDatabaseGetNumberOfPages(struct vfsDatabase *d)
+{
+	uint8_t *page;
+
+	assert(d->n_pages > 0);
+
+	page = d->pages[0];
+
+	/* The page size is stored in the 16th and 17th bytes of the first
+	 * database page (big-endian) */
+	return vfsGet32(&page[28]);
+}
+
 static void vfsDatabaseSnapshot(struct vfsDatabase *d, uint8_t **cursor)
 {
 	uint32_t page_size;
@@ -2378,6 +2431,7 @@ static void vfsDatabaseSnapshot(struct vfsDatabase *d, uint8_t **cursor)
 
 	page_size = vfsDatabaseGetPageSize(d);
 	assert(page_size > 0);
+	assert(d->n_pages == vfsDatabaseGetNumberOfPages(d));
 
 	for (i = 0; i < d->n_pages; i++) {
 		memcpy(*cursor, d->pages[i], page_size);
@@ -2425,6 +2479,10 @@ int VfsSnapshot(sqlite3_vfs *vfs, const char *filename, void **data, size_t *n)
 		return 0;
 	}
 
+	if (database->n_pages != vfsDatabaseGetNumberOfPages(database)) {
+		return SQLITE_CORRUPT;
+	}
+
 	wal = &database->wal;
 
 	*n = vfsDatabaseFileSize(database) + vfsWalFileSize(wal);
